from django.db import models


class RawSugarQuery(models.sql.Query):
    def __init__(self, *args, _source=None, **kwargs):
        self._source = _source
        return super().__init__(*args, **kwargs)

    def get_compiler(self, *args, **kwargs):
        compiler = super().get_compiler(*args, **kwargs)
        get_from_clause_method = compiler.get_from_clause

        def get_from_clause_wrapper(*args, **kwargs):
            qn = compiler.connection.ops.quote_name
            result, params = get_from_clause_method(*args, **kwargs)
            if len(self._source.translations) > 0 or len(self._source.null_fields) > 0:
                wrapper_fields = []
                for left, right in self._source.translations.items():
                    wrapper_fields.append(
                        '{}.{} AS {}'.format(qn('wrapper_table'),  qn(left), qn(right)))
                for null_field in self._source.null_fields:
                    wrapper_fields.append('NULL AS {}'.format(qn(null_field)))
                wrapper = '(SELECT *, {} FROM {} AS {})'.format(
                    ','.join(wrapper_fields),
                    self._source.raw_query,
                    qn('wrapper_table'))
            else:
                wrapper = self._source.raw_query
            result[0] = '{} AS {}'.format(
                wrapper, qn(self.model._meta.db_table))
            params = tuple(self._source.params) + tuple(params)
            return result, params

        compiler.get_from_clause = get_from_clause_wrapper
        return compiler


class RawSugarQuerySet(models.QuerySet):
    def __init__(self, *args, query=None, _source=None, **kwargs) -> None:
        empty_query = query is None
        r = super().__init__(*args, query=query, **kwargs)
        if empty_query:
            self.query = RawSugarQuery(
                self.model,
                _source=_source)
        return r



from django.db import models
from .query import RawSugarQuerySet
from .sources import FromRaw, FromQuerySet


class RawManagerMixin:
    def from_raw(self, raw_query=None, params=[],
                 translations={}, null_fields=[],
                 db_table=None):
        source = FromRaw(raw_query, params, translations,
                         null_fields, db_table)
        return RawSugarQuerySet(self.model,
                                using=self._db,
                                _source=source)

    def from_queryset(self, queryset, translations={}):
        source = FromQuerySet(queryset, translations)
        source._calculate_null_fields(self.model)
        return RawSugarQuerySet(self.model,
                                using=self._db,
                                _source=source)


class RawManager(models.Manager, RawManagerMixin):
    def __init__(self, from_raw=None):
        if from_raw is not None:
            assert isinstance(from_raw, FromRaw), \
                "Expected a `FromRaw` to be passed "\
                "to 'from_raw', but received a `%s" % type(from_raw)
        self._source = from_raw
        return super().__init__()

    def get_queryset(self):
        if isinstance(self._source, FromRaw):
            return RawSugarQuerySet(self.model,
                                    using=self._db,
                                    _source=self._source)
        else:
            return super().get_queryset()


class DecoratedRawManager(models.Manager, RawManagerMixin):
    def __init__(self, *args, _source_func=None, _source_func_is_callable=False,
                 _set_source_func_on_next_call=False, **kwargs):
        self._source_func = _source_func
        if not _source_func_is_callable:
            assert callable(self._source_func)
        self._source_func_is_callable = _source_func_is_callable
        self._set_source_func_on_next_call = _set_source_func_on_next_call
        return super().__init__(*args, **kwargs)

    def _check_source_type(self, source):
        assert isinstance(source, (FromRaw, FromQuerySet)), \
            "Expected a `FromRaw` or `FromQuerySet` to be returned "\
            "from the manager method, but received a `%s" % type(source)

    def __call__(self, *args, **kwargs):
        if self._set_source_func_on_next_call:
            self._source_func = args[0]
            assert callable(self._source_func)
            self._set_source_func_on_next_call = False
            return self
        elif self._source_func_is_callable:
            source = self._source_func(self.model, *args, **kwargs)
            self._check_source_type(source)
            if isinstance(source, FromQuerySet):
                source._calculate_null_fields(self.model)
            return RawSugarQuerySet(self.model,
                                    using=self._db,
                                    _source=source)
        else:
            raise TypeError

    def get_queryset(self):
        if self._source_func_is_callable:
            raise TypeError

        source = self._source_func(self.model)
        self._check_source_type(source)

        return RawSugarQuerySet(self.model,
                                using=self._db,
                                _source=source)


def raw_manager(is_callable=False):
    if callable(is_callable):
        return DecoratedRawManager(_source_func=is_callable)
    if is_callable:
        return DecoratedRawManager(_set_source_func_on_next_call=True,
                                   _source_func_is_callable=True)
    else:
        return DecoratedRawManager(_set_source_func_on_next_call=True)



from django.db import connection


class FromRaw:
    def __init__(self, raw_query=None, params=[],
                 translations={}, null_fields=[],
                 db_table=None):
        assert raw_query or db_table, \
            "Either `raw_query` or `db_table` must be provided!"
        assert not raw_query or not db_table, \
            "Either `raw_query` or `db_table` must be provided, not both!"
        self.raw_query = db_table or '({})'.format(raw_query)
        self.params = params
        self.translations = translations
        self.null_fields = null_fields


class FromQuerySet:
    def __init__(self, queryset, translations={}):
        queryset_fields = list(queryset.query.annotations.keys())
        if translations:
            queryset_fields += [translations[i] for i in translations]
        if len(queryset.query.values_select) > 0:
            for field_name in queryset.query.values_select:
                for field in queryset.model._meta.fields:
                    if field.name == field_name:
                        queryset_fields.append(field.column)
                        break
                else:
                    queryset_fields.append(field_name)
        else:
            queryset_fields += [field.column for field in queryset.model._meta.fields]
        self._queryset_fields = queryset_fields

        raw_query, params = queryset.query.as_sql(
            connection=connection, compiler=None)
        self.raw_query = "({})".format(raw_query)
        self.params = params
        self.translations = translations
        self.null_fields = []

    def _calculate_null_fields(self, target_model):
        model_fields = [f.column for f in target_model._meta.fields]
        self.null_fields = list(set(model_fields) - set(self._queryset_fields))




DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': ':memory:',
    },
}

INSTALLED_APPS = (
    'django.contrib.contenttypes',
    'tests',
)

SECRET_KEY = 'foobar'



from django.test import TestCase
from .models import MySimpleModel, AnotherSimpleModel


class DynamicQueringTest(TestCase):
    def setUp(self):
        asm1 = AnotherSimpleModel.objects.create()
        asm2 = AnotherSimpleModel.objects.create()
        msm1 = MySimpleModel.objects.create(
            name='django model 1', number=1, source=asm1)
        msm2 = MySimpleModel.objects.create(
            name='django model 2', number=2, source=asm2)

    def test_readme_example_0(self):
        queryset = MySimpleModel.objects.all().order_by('id')

        self.assertTrue(queryset.count() == 2)

        res = queryset[0]

        self.assertTrue(res.name == 'django model 1')
        self.assertTrue(res.number == 1)
        self.assertTrue(res.source_id == 1)

    def test_readme_example_1(self):
        queryset = MySimpleModel.objects.from_raw(
            'SELECT Null as id, "my str" as name, 111 as number, Null as source_id')

        queryset = queryset.filter(number__gte=10)\
            .exclude(number__gte=1000)\
            .filter(name__contains='s')\
            .order_by('number')\
            .select_related('source')

        res = queryset[0]

        self.assertTrue(res.name == 'my str')
        self.assertTrue(res.number == 111)

    def test_readme_example_2(self):
        queryset = MySimpleModel.objects.from_raw(
            'SELECT "my str" as name, 111 as number', null_fields=['id', 'source_id'])

        queryset = queryset.filter(number__gte=10)\
            .exclude(number__gte=1000)\
            .filter(name__contains='s')\
            .order_by('number')\
            .select_related('source')

        res = queryset[0]

        self.assertTrue(res.name == 'my str')
        self.assertTrue(res.number == 111)

    def test_readme_example_3(self):
        queryset = MySimpleModel.objects.from_raw(
            'SELECT %s as name, 111 as number',
            params=['my str'],
            null_fields=['id', 'source_id'])

        queryset = queryset.filter(number__gte=10)\
            .exclude(number__gte=1000)\
            .filter(name__contains='s')\
            .order_by('number')

        res = queryset[0]

        self.assertTrue(res.name == 'my str')
        self.assertTrue(res.number == 111)

    def test_readme_example_4(self):
        queryset = MySimpleModel.objects.from_raw(
            'SELECT %s as name, 111 as inner_number',
            params=['my str'],
            translations={'inner_number': 'number'},
            null_fields=['id', 'source_id'])

        queryset = queryset.filter(number__gte=10)\
            .exclude(number__gte=1000)\
            .filter(name__contains='s')\
            .order_by('number')

        res = queryset[0]

        self.assertTrue(res.name == 'my str')
        self.assertTrue(res.number == 111)

    def test_readme_example_5(self):
        queryset = MySimpleModel.my_raw_manager.all()

        queryset = queryset.filter(number__gte=10)\
            .exclude(number__gte=1000)\
            .filter(name__contains='s')\
            .order_by('number')

        res = queryset[0]

        self.assertTrue(res.name == 'my str')
        self.assertTrue(res.number == 111)

    def test_readme_example_6(self):
        queryset = MySimpleModel.my_raw_manager_2.all()

        queryset = queryset.filter(number__gte=10)\
            .exclude(number__gte=1000)\
            .filter(name__contains='s')\
            .order_by('number')

        res = queryset[0]

        self.assertTrue(res.name == 'my str')
        self.assertTrue(res.number == 111)

    def test_readme_example_7(self):
        queryset = MySimpleModel.my_callable_raw_manager('my str').all()

        queryset = queryset.filter(number__gte=10)\
            .exclude(number__gte=1000)\
            .filter(name__contains='s')\
            .order_by('number')

        res = queryset[0]

        self.assertTrue(res.name == 'my str')
        self.assertTrue(res.number == 111)

    def test_decorated_manager_from_1(self):
        queryset = MySimpleModel.my_raw_manager_2.from_raw(
            'SELECT %s as name, 111 as inner_number',
            params=['my str'],
            translations={'inner_number': 'number'},
            null_fields=['id', 'source_id'])

        queryset = queryset.filter(number__gte=10)\
            .exclude(number__gte=1000)\
            .filter(name__contains='s')\
            .order_by('number')

        res = queryset[0]

        self.assertTrue(res.name == 'my str')
        self.assertTrue(res.number == 111)

    def test_decorated_manager_from_2(self):
        queryset = MySimpleModel.my_callable_raw_manager.from_raw(
            'SELECT %s as name, 111 as inner_number',
            params=['my str'],
            translations={'inner_number': 'number'},
            null_fields=['id', 'source_id'])

        queryset = queryset.filter(number__gte=10)\
            .exclude(number__gte=1000)\
            .filter(name__contains='s')\
            .order_by('number')

        res = queryset[0]

        self.assertTrue(res.name == 'my str')
        self.assertTrue(res.number == 111)

    def test_call_not_callable(self):
        try:
            MySimpleModel.my_raw_manager_2().all()
        except TypeError:
            return

        self.assertTrue(False)

    def test_query_callable_directly(self):
        try:
            MySimpleModel.my_callable_raw_manager.all()
        except TypeError:
            return

        self.assertTrue(False)
